"""The z_piezo module controls the one-dimensional displacement of a
piezoelectric actuator, while allowing simultaneous data aquisition on
two channels.  The module is called z_piezo, because for force spectroscopy,
the axis that needs simultaneous aquisition is the Z axis, although this
module could be used to provide control in any "fast" axis.

There are a few globals controlling the behavior of the entire module.

  USE_ABCD_DEFLECTION (default False)
    selects between using a preprocessed vertical deflection signal
    and using the 'raw' 4 photodiode output segments.
  TEXT_VERBOSE  (default False)
    print text messages to stderr displaying actions taken
  PYLAB_INTERACTIVE_VERBOSE (default False)
    display pylab graphs of any ramps taken
  BASE_FIG_NUM (default 50)
    the figure number for z_piezo pylab figures will be this number + {0,1,..}
  LOG_RAMPS (default False)
    save the input and output of any ramps to the LOG_DIR directory
  LOG_DIR = '${DEFAULT}/z_piezo'
    where the ramp logs are saved to
  DEFAULT_ZP_CHAN (default 0)
    output channel controlling the z piezo
  DEFAULT_ZP_MON_CHAN (default 1)
    input channel monitoring the z piezo signal
  DEFAULT_DEF_CHAN (default 0, or in ABCD mode (2,3,4,5))
    input channel(s) monitoring deflection (or some other feedback signal)
"""

USE_ABCD_DEFLECTION = False
TEXT_VERBOSE = False
PYLAB_INTERACTIVE_VERBOSE = False
BASE_FIG_NUM = 50
LOG_RAMPS = False
LOG_DIR = '${DEFAULT}/z_piezo'

# Hackish defaults for the calibcant.config module
DEFAULT_GAIN = 20 # Vpiezo / Voutput
DEFAULT_SENSITIVITY = 6.790 # nm_surface/Volt_piezo (for S/N 4546EV piezo from 2009/01/09 calibration on 200nm deep pits, 10 um pitch Digital Instruments P/N 498-000-026)
DEFAULT_ZERO_PHOTODIODE_BITS = 2**15
DEFAULT_VPHOTO_IN_2_VOLTS = lambda vbits : (vbits-DEFAULT_ZERO_PHOTODIODE_BITS)/3276.8
DEFAULT_VZP_OUT_2_VOLTS   = lambda vbits : (vbits-DEFAULT_ZERO_PHOTODIODE_BITS)/3276.8

DEFAULT_ZP_CHAN = 0
DEFAULT_ZP_MON_CHAN = 1
DEFAULT_DEF_CHAN = 0
if USE_ABCD_DEFLECTION :
    DEFAULT_DEF_CHAN = (2,3,4,5)
    # The 4 photodiode segments are
    #  A B
    #  C D
    # looking in along the laser

class NoComedi (ImportError):
    "Missing comedi.  Don't initialize a z_piezo instance."
    pass

try:
    import pycomedi.single_aio as single_aio
    import pycomedi.simult_aio as simult_aio
    HAS_COMEDI = True
except ImportError, e:
    HAS_COMEDI = False

from numpy import array, arange, ones, zeros, linspace, uint16, float, sin, pi
from scipy.stats import linregress
import data_logger

if PYLAB_INTERACTIVE_VERBOSE :
    from pylab import figure, plot, title, legend, hold, subplot
    import time # for timestamping lines on plots

class error (Exception) :
    pass
class outOfRange (error) :
    pass

# frontend:

class z_piezo :
    def __init__(self, zp_chan=DEFAULT_ZP_CHAN,
                 zp_mon_chan=DEFAULT_ZP_MON_CHAN,
                 def_chan=DEFAULT_DEF_CHAN) :
        if HAS_COMEDI is not True:
            raise NoComedi
        self.sensitivity = DEFAULT_SENSITIVITY
        self.gain = DEFAULT_GAIN
        self.nm2Vout = self.sensitivity * self.gain # nm_surface / V_output
        self.chan_info = _chan_info(zp_chan, zp_mon_chan, def_chan)

        self.curIn = array([0]*self.chan_info.numIn, dtype=uint16)
        self.curOut = array([0]*self.chan_info.numOut, dtype=uint16)
        self._jump = _z_piezo_jump(self.chan_info)
        self._ramp = _z_piezo_ramp(self.chan_info)
        self.zpMax = int(self.pos_nm2out(1000)) # limits at 1 micron
        self.zpMin = int(self.pos_nm2out(-1000)) # and -1 micron
        # define default value changed callbacks
        self.setExternalZPiezo = None
        self.setExternalZPiezoMon = None
        self.setExternalDeflection = None
    def __del__(self) :
        pass
    def curPos(self) :
        return int(self.curOut[self.chan_info.zp_ind]) # cast as int so arithmetic is signed
    def curPosMon(self) :
        return int(self.curIn[self.chan_info.zp_mon_ind])
    def curDef(self) :
        if USE_ABCD_DEFLECTION :
            A = int(self.curIn[self.chan_info.def_ind[0]])
            B = int(self.curIn[self.chan_info.def_ind[1]])
            C = int(self.curIn[self.chan_info.def_ind[2]])
            D = int(self.curIn[self.chan_info.def_ind[3]])
            df = float((A+B)-(C+D))/(A+B+C+D)
            dfout = int(df * 2**15) + 2**15
            if TEXT_VERBOSE :
                print "Current deflection %d (%d, %d, %d, %d)" \
                    % (dfout, A, B, C, D)
            return dfout
        else : # reading the deflection voltage directly
            return int(self.curIn[self.chan_info.def_ind])
    def pos_out2V(self, out) :
        return self._jump.out_to_phys(out)
    def pos_V2out(self, V) :
        # Vout is Volts output by the DAQ card
        # and Vpiezo is Volts applied across the piezo
        return self._jump.phys_to_out(V)
    def pos_nm2out(self, nm) :
        # Vout is Volts output by the DAQ card
        # and Vpiezo is Volts applied across the piezo
        return self.pos_V2out(nm / self.nm2Vout)
    def pos_out2nm(self, out) :
        return self.pos_out2V(out) * self.nm2Vout
    def posMon_nm2out(self, nm) :
        return self._jump.phys_to_out(nm / self.nm2Vout)
    def posMon_out2nm(self, out) :
        return self._jump.out_to_phys(out) * self.nm2Vout
    def def_in2V(self, input) :
        return self._jump.out_to_phys(input)
    def def_V2in(self, physical) :
        print physical, type(physical) # temporary debugging printout
        return self._jump.phys_to_out(physical)
    def curVals(self) :
        return {'Z piezo output': self.curPos(),
                'Deflection input':self.curDef(),
                'Z piezo input':self.curPosMon()}
    def pCurVals(self, curVals=None) :
        if curVals == None :
            curVals = self.curVals()
        print "Z piezo output  : %6d" % curVals["Z piezo output"]
        print "Z piezo input   : %6d" % curVals["Z piezo input"]
        print "Deflection input: %6d" % curVals["Deflection input"]
    def updateInputs(self) :
        self.curIn = self._jump.updateInputs()
        if self.setExternalZPiezoMon != None :
            self.setExternalZPiezoMon(self.posMon_out2nm(self.curPosMon()))
        if self.setExternalDeflection != None :
            self.setExternalDeflection(self.def_in2V(self.curDef()))
    def jumpToPos(self, pos) :
        self._check_range(pos)
        self.curOut[self.chan_info.zp_ind] = pos
        self.curIn = self._jump.jumpToPos(self.curOut)
        if self.setExternalZPiezo != None :
            self.setExternalZPiezo(self.pos_out2nm(self.curPos()))
        if TEXT_VERBOSE :
            print "Jumped to"
            self.pCurVals()
        return self.curVals()
    def ramp(self, out, freq) :
        for pos in out :
            self._check_range(pos)
        out = self._ramp.ramp(out, freq)
        self.curOut[self.chan_info.zp_ind] = out["Z piezo output"][-1]
        self.curIn[self.chan_info.zp_mon_ind] = out["Z piezo input"][-1]
        if USE_ABCD_DEFLECTION :
            for i in range(4) : # i is the photodiode element (0->A, 1->B, ...)
                self.curIn[i] = out["Deflection segment"][i][-1]
        else :
            self.curIn[self.chan_info.def_ind] = out["Deflection input"][-1]

        if self.setExternalZPiezo != None :
            self.setExternalZPiezo(self.pos_out2nm(self.curPos()))
        if self.setExternalZPiezoMon != None :
            self.setExternalZPiezoMon(self.posMon_out2nm(self.curPosMon()))
        if self.setExternalDeflection != None :
            self.setExternalDeflection(self.def_in2V(self.curDef()))
        if LOG_RAMPS :
            log = data_logger.data_log(LOG_DIR, noclobber_logsubdir=False,
                                       log_name="ramp")
            if USE_ABCD_DEFLECTION :
                v = out.pop("Deflection segment")
                log.write_dict_of_arrays(out)
                out["Deflection segment"] = v
            else :
                log.write_dict_of_arrays(out)
        if TEXT_VERBOSE :
            print "Ramped from (%g, %g) to (%g, %g) in %d points" \
                % (out["Z piezo output"][0], out["Deflection input"][0],
                   out["Z piezo output"][-1], out["Deflection input"][-1],
                   len(out["Z piezo output"]))
        if PYLAB_INTERACTIVE_VERBOSE :
            figure(BASE_FIG_NUM)
            subplot(211)
            title('ramp (def_in vs zp_out)')
            timestamp = time.strftime('%H%M%S')
            plot(out["Z piezo output"], out["Deflection input"], '.', label=timestamp)
            subplot(212)
            title('zp_in vs zp_out')
            plot(out["Z piezo output"], out["Z piezo input"], '.', label=timestamp)
            legend(loc='best')
        return out
    def _check_range(self, pos) :
        if pos > self.zpMax :
            raise outOfRange, "Z piezo pos = %d > %d = max" % (pos, self.zpMax)
        if pos < self.zpMin :
            raise outOfRange, "Z piezo pos = %d < %d = min" % (pos, self.zpMin)

# backends:
 
class _z_piezo_jump :
    def __init__(self, chan_info) :
        self.chan_info = chan_info
        self.reserved = False
        self.AI = single_aio.AI(chan=self.chan_info.inChan)
        self.AI.close()
        self.AO = single_aio.AO(chan=self.chan_info.outChan)
        self.AO.close()
    def reserve(self) :
        self.AI.open()
        self.AO.open()
        self.reserved = True
    def release(self) :
        self.AI.close()
        self.AO.close()
        self.reserved = False
    def updateInputs(self) :
        if self.reserved == False :
            self.AI.open()
            #self.reserve()
            self.reserved = False
        In = self.AI.read()
        if self.reserved == False :
            self.AI.close()
            #self.release()
        return In
    def jumpToPos(self, out) :
        if self.reserved == False :
            self.reserve()
            self.reserved = False # set to true inside reserve(), which we don't want
        self.AO.write(out)
        if self.reserved == False :
            self.release()
        return self.updateInputs()
    def out_to_phys(self, output) :
        return self.AO.comedi_to_phys(0, output)
    def phys_to_out(self, physical) :
        return self.AO.phys_to_comedi(0, physical)

class _z_piezo_ramp :
    def __init__(self, chan_info) :
        self.chan_info = chan_info
        self.attempts=0
        self.failures=0
        self.verbose = True
    def ramp(self, outArray, freq) :
        # allocate and run the task
        npoints = int(len(outArray)/self.chan_info.numOut)
        in_data = array([0]*npoints*self.chan_info.numIn, dtype=uint16)
        if type(outArray) != type(in_data) or outArray.dtype != uint16 :
            out_data = array([0]*npoints*self.chan_info.numOut, dtype=uint16)
            for i in range(npoints) :
                out_data[i] = outArray[i]
        else :
            out_data = outArray
        if simult_aio.DONT_OUTPUT_LAST_SAMPLE_HACK == True:
            # duplicate the last output point
            npoints += 1
            out_hack = array([0]*npoints*self.chan_info.numOut, dtype=uint16)
            for i in range(npoints-1) :
                out_hack[i] = out_data[i]
            out_hack[-1] = out_data[-1]
            out_data = out_hack
            in_data = array([0]*npoints*self.chan_info.numIn, dtype=uint16)

        correlated = False
        ramp={}
        while not correlated :
            AIO = simult_aio.AIO(in_chan=self.chan_info.inChan,
                                 out_chan=self.chan_info.outChan)
            if self.verbose :
                print "setup AIO"
            AIO.setup(freq=freq, out_buffer=out_data)
            if self.verbose :
                print "arm AIO"
            AIO.arm()
            if self.verbose :
                print "read AIO"
            AIO.start_read(in_data)
            if self.verbose :
                print "close AIO"
            AIO.close()
            if self.verbose :
                print "finished AIO"
            self.attempts += 1
            ramp["Z piezo output"] = out_data[self.chan_info.zp_ind::self.chan_info.numOut]
            ramp["Z piezo input"] = in_data[self.chan_info.zp_mon_ind::self.chan_info.numIn]
            ramp["Deflection input"] = in_data[self.chan_info.def_ind::self.chan_info.numIn]
            failed = False
            gradient, intercept, r_value, p_value, std_err = linregress(ramp["Z piezo output"],
                                                                        ramp["Z piezo input"])
            rnge = ramp["Z piezo output"].max()-ramp["Z piezo output"].min()
            if gradient < .8 and rnge > 100 :
                if PYLAB_INTERACTIVE_VERBOSE :
                    figure(BASE_FIG_NUM+3)
                    subplot(211)
                    title('ramp (def_in vs zp_out)')
                    timestamp = time.strftime('%H%M%S')
                    plot(ramp["Z piezo output"], ramp["Deflection input"], '.', label=timestamp)
                    subplot(212)
                    title('zp_in vs zp_out')
                    plot(ramp["Z piezo output"], ramp["Z piezo input"], '.', label=timestamp)
                    legend(loc='best')
                print "ramp failed slope (%g, %g), try again" % (gradient, rnge)
                failed = True
            if not failed :
                if simult_aio.DONT_OUTPUT_LAST_SAMPLE_HACK == True:
                    # remove the duplicated last output point
                    out_data = out_data[:-self.chan_info.numOut]
                    in_data = in_data[:-self.chan_info.numIn]
                ramp["Z piezo output"] = out_data[self.chan_info.zp_ind::self.chan_info.numOut]
                ramp["Z piezo input"] = in_data[self.chan_info.zp_mon_ind::self.chan_info.numIn]
                ramp["Deflection input"] = in_data[self.chan_info.def_ind::self.chan_info.numIn]

                if USE_ABCD_DEFLECTION :
                    ramp["Deflection segment"] = []
                    for c_ind in self.chan_info.def_ind :
                        ramp["Deflection segment"].append(in_data[c_ind::self.chan_info.numIn])
                    A = ramp["Deflection segment"][0].astype(float)
                    B = ramp["Deflection segment"][1].astype(float)
                    C = ramp["Deflection segment"][2].astype(float)
                    D = ramp["Deflection segment"][3].astype(float)
                    ramp["Deflection input"] = (((A+B)-(C+D))/(A+B+C+D) * 2**15).astype(uint16)
                    ramp["Deflection input"] += 2**15 # HACK, comedi uses unsigned ints...
                else :
                    ramp["Deflection input"] = in_data[self.chan_info.def_ind::self.chan_info.numIn]
                correlated = True
                if PYLAB_INTERACTIVE_VERBOSE :
                    figure(BASE_FIG_NUM+1)
                    timestamp = time.strftime('%H%M%S')
                    plot(ramp["Z piezo output"], ramp["Z piezo input"], '.', label=timestamp)
                    title('goods')
                    legend(loc='best')
            else :
                self.failures += 1
                if PYLAB_INTERACTIVE_VERBOSE :
                    figure(BASE_FIG_NUM+2)
                    timestamp = time.strftime('%H%M%S')
                    plot(ramp["Z piezo output"], ramp["Z piezo input"], '.', label=timestamp)
                    legend(loc='best')
                    title('bads')

        #AIO = simAIO.simAIOobj(npoints,
        #                       freq,
        #                       self.chan_info.inChan,
        #                       self.chan_info.outChan)
        #err = AIO.configure()
        #err = AIO.commit()
        #err = AIO.write(outArray)
        #err = AIO.run()
        #(err, inArray) = AIO.read()
        #if err != 0 :
        #    raise error, "Error in simAIO"
        #if len(inArray)*self.chan_info.numOut != len(outArray)*self.chan_info.numIn :
        #    raise error, "Input array of wrong length"
        #
        #ramp={}
        #ramp["Z piezo output"] = outArray[self.chan_info.zp_ind::self.chan_info.numOut]
        #ramp["Z piezo input"] = inArray[self.chan_info.zp_mon_ind::self.chan_info.numIn]
        #ramp["Deflection input"] = inArray[self.chan_info.def_ind::self.chan_info.numIn]
        
        # task automatically freed as AIO goes out of scope
        return ramp

class _chan_info :
    def __init__(self, zp_chan, zp_mon_chan, def_chan) :
        self.inChan = []
        if USE_ABCD_DEFLECTION :
            self.def_ind = [] # index in inChan array
            i=0
            for chan in def_chan :
                self.inChan.append(chan)
                self.def_ind.append(i)
                i += 1
        else :
            self.inChan.append(def_chan)
            self.def_ind = 0 # index in inChan array
        self.inChan.append(zp_mon_chan)
        self.zp_mon_ind = zp_mon_chan
        self.numIn = len(self.inChan)
        
        self.outChan = [zp_chan]
        self.zp_ind = 0 # index in outChan array
        self.numOut = len(self.outChan)
        

def test() :
    print "test z_piezo()"
    zp = z_piezo()
    zp.verbose = True
    zp.jumpToPos(zp.pos_nm2out(0))
    if zp.verbose :
        zp.pCurVals()
    curPos = zp.curPos()
    curPosNm = zp.pos_out2nm(curPos)
    curPos2 = zp.pos_nm2out(curPosNm)
    if curPos != curPos2 :
        raise error, "Conversion %d -> %g -> %d not round trip" % (curPos, curPosNm, curPos2)
    ramp_positions = linspace(zp.curPos(), zp.pos_nm2out(100), 20)
    out = zp.ramp(ramp_positions, 10000)
    ramp_positions = linspace(zp.curPos(), zp.pos_nm2out(0), 20)
    out2 = zp.ramp(ramp_positions, 10000)

if __name__ == "__main__" :
    test()
